#include "Obstacles.h"
#include <stdio.h>

Obstacle::Obstacle(int type, float x, float z, float step) {
   this->type = type;
   this->x = x;
   this->z = z;
   this->step = step;
}

float Obstacle::move() {
   z += step;
   return z;
}

Particle::Particle(vec3 pos, vec3 force, float gravity) {
   this->pos = pos;
   this->force = force;
   this->gravity = gravity;
   steps = 0;

   this->Dcolor = vec3((float)rand() / (float)RAND_MAX * 0.3,
                       (float)rand() / (float)RAND_MAX * 0.3,
                       (float)rand() / (float)RAND_MAX * 0.3 + 0.7);

   this->Scolor = vec3((float)rand() / (float)RAND_MAX * 0.1,
                       (float)rand() / (float)RAND_MAX * 0.1,
                       (float)rand() / (float)RAND_MAX * 0.3);

   this->Acolor = vec3(0.1, 0.1, 0.2);
   this->shine = (float)rand() / (float)RAND_MAX * 0.3;
}

Particle::Particle(vec3 pos, vec3 force, float gravity, vec3 diff, vec3 spec, vec3 amb, float shine) {
   this->pos = pos;
   this->force = force;
   this->gravity = gravity;
   Dcolor = diff;
   Scolor = spec;
   Acolor = amb;
   shine = shine;
   steps = 0;
}

float Particle::move() {
   pos += force;
   pos.y -= steps * gravity;
   steps++;

   return pos.y;
}

Pole::Pole(int type, float x, float z, float step): Obstacle(type, x, z, step) {}

Wall::Wall(int type, float x, float z, float step, float width) : Obstacle(type, x, z, step) {
   this->width = width;
}

Spike::Spike(int type, float x, float z, float step, int num) : Obstacle(type, x, z, step) {
   this->num = num;
}

ObstacleList::Node::Node(int type, float x, float z, float step, float width) {
   switch (type) {
      case POLE:
         obstacle = new Pole(type, x, z, step);
         break;
      case WALL:
         obstacle = new Wall(type, x, z, step, width);
         break;
      case SPIKE:
         obstacle = new Spike(type, x, z, step, (int)width);
         break;
   }
}

ObstacleList::ObstacleList()
   : size(0) {}

void ObstacleList::add(int type, float x, float z, float step, float width) {
   Node *newNode = new Node(type, x, z, step, width), *temp;
   newNode->next = newNode->prev = NULL;
   Obstacle *o;
   int i;

   if (!size) {
      head = tail = newNode;
   }
   else {
      temp = head;
      for (i = 0; i < size; i++) {
         o = temp->obstacle;

         if (newNode->obstacle->z > o->z) {
            if (!i) {
               head->prev = newNode;
               newNode->next = head;
               head = newNode;
            } else {
               temp->prev->next = newNode;
               newNode->prev = temp->prev;
               temp->prev = newNode;
               newNode->next = temp;
            }
            break;
         }
         temp = temp->next;
      }
      if (i == size) {
         tail->next = newNode;
         newNode->prev = tail;
         tail = newNode;
      }
   }
   
   size++;
}

void ObstacleList::remove() {
   assert(size > 0);

   Node *temp = head;
   head = head->next;
      
   if (size == 1) {
      tail = head;
   }
   else {
      head->prev = NULL;
   }

   temp->~Node();
   size--;
}

void ObstacleList::remove(int index) {
   assert(size > 0);
   assert(index >= 0 && index < size);
   Node *temp = head;
      
   if (size == 1) {
      head = tail = NULL;
   } else {
      for (int i = 0; i < index; i++) {
         temp = temp->next;
      }

      if (temp->prev) {
         temp->prev->next = temp->next;
      } else {
         head = head->next;
         head->prev = NULL;
      }

      if (temp->next) {
         temp->next->prev = temp->prev;
      } else {
         tail = tail->prev;
         tail->next = NULL;
      }
   }

   temp->~Node();
   size--;
}

void ObstacleList::move() {
   Node *temp = head;
   int toRemove = 0;

   for (int i = 0; i < size; i++) {
      if (temp->obstacle->move() > 5) {
         toRemove++;
      }
      temp = temp->next;
   }

   while (toRemove-- > 0) {
      remove();
   }
}

void ObstacleList::draw(void (*drawPole)(float, float, float), void (*drawWall)(float, float, float, float), void (*drawSpikes)(float, float, float, int)) {
   Node *temp = head;
   Obstacle *o;

   for (int i = 0; i < size; i++) {
      o = temp->obstacle;
      switch (o->type) {
         case POLE:
            drawPole(o->x, 0, o->z);
            break;
         case WALL:
            drawWall(o->x, 0, o->z, ((Wall*)o)->width);
            break;
         case SPIKE:
            drawSpikes(o->x, 0, o->z, ((Spike*)o)->num);
            break;
      }

      temp = temp->next;
   }
}

bool ObstacleList::isColliding(float runnerX, float runnerY, float width, float length, bool isPunching, ParticleList *pl) {
   Node *temp = head;
   Obstacle *o;
   bool collide = false;
   int particles = 10;
   vec3 pos, force;
   float gravity = 0.02;
   float punchHeight = 3;

   //printf("size: %d\n", size);
   for (int i = 0; i < size && !collide; i++) {
      o = temp->obstacle;
      switch (o->type) {
         case POLE:
            if (runnerX + width > o->x - 0.25 && runnerX - width < o->x + 0.25 && 
                -length < o->z + 0.25 && length > o->z - 0.25) {
               collide = true;
               //printf("Collided with POLE\n");
            }
            break;
         case WALL:
            if (runnerX + width > o->x - ((Wall*)o)->width / 2 && runnerX - width < o->x + ((Wall*)o)->width / 2  && 
                -length < o->z + 0.25 && length > o->z - 0.25) {
               collide = true;
               //printf("Collided with WALL\n");
               
               if (isPunching) {
                  //printf("Too bad we F#$%ED THAT WALL UP! deleting object:%d\n", i);
                  collide = false;
                  while (particles--) {
                     pos = vec3(runnerX + randRange(.25), punchHeight + rand1() * .25, o->z + -rand1() * 0.25 + 0.25);
                     force = vec3(randRange(0.25), rand1() * 0.4 + .3, -rand1() * .2 - .25);
                     pl->add(pos, force, gravity);
                  }
                  remove(i);

                  return collide;
               }
            }
            break;
         case SPIKE:
            if (runnerX + width > o->x - (float)((Spike*)o)->num / 2 + 0.2 && runnerX - width < o->x + (float)((Spike*)o)->num / 2 - 0.2 &&
                -length < o->z + 0.35 && length > o->z - 0.35 &&
                runnerY < 0.8) {
               collide = true;
               //printf("Collided with SPIKE\n");
            }
            break;
      }

      temp = temp->next;
   }

   return collide;
}

ParticleList::ParticleList() 
   : size(0) {}

ParticleList::Node::Node(vec3 pos, vec3 force, float gravity) {
   this->particle = new Particle(pos, force, gravity);
}

ParticleList::Node::Node(vec3 pos, vec3 force, float gravity, vec3 diff, vec3 spec, vec3 amb, float shine) {
   this->particle = new Particle(pos, force, gravity, diff, spec, amb, shine);
}

void ParticleList::add(vec3 pos, vec3 force, float gravity) {
   Node *newNode = new Node(pos, force, gravity);
   newNode->next = newNode->prev = NULL;
   Particle *p;
   int i;

   if (!size) {
      head = tail = newNode;
   }
   else {
      head->prev = newNode;
      newNode->next = head;
      head = newNode;
   }
   
   size++;
}

void ParticleList::add(vec3 pos, vec3 force, float gravity, vec3 diff, vec3 spec, vec3 amb, float shine) {
   Node *newNode = new Node(pos, force, gravity, diff, spec, amb, shine);
   newNode->next = newNode->prev = NULL;
   Particle *p;
   int i;

   if (!size) {
      head = tail = newNode;
   }
   else {
      head->prev = newNode;
      newNode->next = head;
      head = newNode;
   }
   
   size++;
}

void ParticleList::remove() {
   assert(size > 0);

   Node *temp = head;
   head = head->next;
      
   if (size == 1) {
      tail = head;
   }
   else {
      head->prev = NULL;
   }

   temp->~Node();
   size--;
}

void ParticleList::remove(int index) {
   assert(size > 0);
   assert(index >= 0 && index < size);
   Node *temp = head;
      
   if (size == 1) {
      head = tail = NULL;
   } else {
      for (int i = 0; i < index; i++) {
         temp = temp->next;
      }

      if (temp->prev) {
         temp->prev->next = temp->next;
      } else {
         head = head->next;
         head->prev = NULL;
      }

      if (temp->next) {
         temp->next->prev = temp->prev;
      } else {
         tail = tail->prev;
         tail->next = NULL;
      }
   }

   temp->~Node();
   size--;
}

void ParticleList::deleteUnder(float height) {
   Node *temp = head, *deleteThis;
   Particle *p;

   for (int i = 0; i < size; i++) {
      p = temp->particle;

      if (p->pos.y < height) {
         if (temp->next) {
            temp->next->prev = temp->prev;
         }
         if (temp->prev) {
            temp->prev->next = temp->next;
         }
         deleteThis = temp;
         temp = temp->next;
         deleteThis->~Node();
         size--;
      } else {
         temp = temp->next;
      }
   }
}

void ParticleList::move() {
   Node *temp = head;

   for (int i = 0; i < size; i++) {
      temp->particle->move();
      temp = temp->next;
   }

   deleteUnder(-15);
}

void ParticleList::draw(void (*drawParticle)(float, float, float, vec3, vec3, vec3, float)) {
   Node *temp = head;
   Particle *p;

   for (int i = 0; i < size; i++) {
      p = temp->particle;
      drawParticle(p->pos.x, p->pos.y, p->pos.z, p->Dcolor, p->Scolor, p->Acolor, p->shine);
      temp = temp->next;
   }
}
