#ifndef Obstacles_H_
#define Obstacles_H_

#include <assert.h>
#include <cstddef>
#include <cstdlib>
#include "../glm/glm.hpp"

#define POLE 1
#define WALL 2
#define SPIKE 3

#define rand1() ((float)rand()/(float)RAND_MAX)
#define randRange(a) ((float)rand()/(float)RAND_MAX*(a*2)-a)

using namespace glm;

class Obstacle {
   public:
      int type;
      float x, z, step;

      Obstacle(int type, float x, float z, float step);
      float move();
};


class Particle {
   public:
      vec3 pos;
      vec3 force;
      float gravity;
      vec3 Dcolor;
      vec3 Scolor;
      vec3 Acolor;
      float shine;
      int steps;

      Particle(vec3 pos, vec3 force, float gravity);
      Particle(vec3 pos, vec3 force, float gravity, vec3, vec3, vec3, float);
      float move();
};


class Pole: public Obstacle {
   public:
      Pole(int type, float x, float z, float step);
};


class Wall: public Obstacle {
   public:
      float width;
      bool alive;

      Wall(int type, float x, float z, float step, float width);
};


class Spike: public Obstacle {
   public:
      int num;

      Spike(int type, float x, float z, float step, int num);
};


class ParticleList {
   public:
      class Node {
         public:
            Node *next, *prev;
            Particle *particle;

            Node(vec3 pos, vec3 force, float gravity);
            Node(vec3 pos, vec3 force, float gravity, vec3, vec3, vec3, float);
      };

      Node *head, *tail;
      int size;

      ParticleList();

      void add(vec3 pos, vec3 force, float gravity);
      void add(vec3 pos, vec3 force, float gravity, vec3 diff, vec3 spec, vec3 amb, float shine);
      void remove();
      void remove(int index);
      void deleteUnder(float height);
      void move();
      void draw(void (*drawParticle)(float, float, float, vec3, vec3, vec3, float));
};


class ObstacleList {
   public:
      class Node {
         public:
            Node *next, *prev;
            Obstacle *obstacle;

            Node(int type, float x, float z, float step, float width);
      };

      Node *head, *tail;
      int size;

      ObstacleList();

      void add(int type, float x, float z, float step, float width);
      void remove();
      void remove(int index);
      void move();
      void draw(void (*drawPole)(float, float, float),
                void (*drawWall)(float, float, float, float),
                void (*drawSpikes)(float, float, float, int));
      bool isColliding(float runnerX, float runnerY, float width, float length, bool isPunching, ParticleList *pl);
};

#endif
