#include <iostream>

#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif

#ifdef __unix__
#include <GL/glut.h>
#endif

#ifdef _WIN32
#pragma comment(lib, "glew32.lib")

#include <GL\glew.h>
#include <GL\glut.h>
#endif

#include "CMeshLoaderSimple.h"
#include <stdlib.h>
#include <stdio.h>
#include <cstdlib>
#include "GLSL_helper.h"
#include "MStackHelp.h"
#include "../glm/glm.hpp"
#include "../glm/gtc/matrix_transform.hpp"
#include "../glm/gtc/type_ptr.hpp"
#include "Obstacles.h"

void tick();

typedef struct Image {
  unsigned long sizeX;
  unsigned long sizeY;
  char *data;
} Image;

int ImageLoad(char *filename, Image *image);
GLvoid LoadTexture(char* image_file, int tex_id);

int materialID = 1;

using namespace std;
using namespace glm;

// Data handles
GLuint GrndBuffObj;

// Shader stuff and window size
int shade = 1;
int ShadeProg;
static float g_width, g_height;

// Camera position
vec3 fixedCam = vec3(0, 8, 7);
vec3 sideCam = vec3(-4.4, 1.4, -1);
vec3 g_camPos = fixedCam;
float pi = 3.14159265359;
float alpha = 2 * pi / 3, beta = -pi / 2.0; // pitch and yaw in radians
//float alpha = pi / 2, beta = pi / 20; // for sideview

// Lookat position
vec3 g_lookPos = vec3(g_camPos.x + cos(beta) * sin(alpha),
                      g_camPos.y + cos(alpha),
                      g_camPos.z + sin(beta) * sin(alpha));
int camMethod = 0; // Default fixed

int g_mouseX, g_mouseY; // mouse position
vec3 g_gazeVec = normalize(g_lookPos - g_camPos);
vec3 up = vec3(0.0, 1.0, 0.0);
vec3 g_vVec, g_wVec = -g_gazeVec;
vec3 g_uVec = normalize(cross(up, g_wVec));

vec3 dirLight = normalize(vec3(0.0, 1.0, -0.1));
int g_mat_id = 0; // Material ID
int lightMethod = 1; // Default Phong lighting

// Handles
GLint h_aPosition;
GLint h_aNormal;

GLint h_uModelMatrix;
GLint h_uViewMatrix;
GLint h_uProjMatrix;

GLint h_viewPos;
GLint h_uMatDif;
GLint h_uMatSpec;
GLint h_uMatAmb;
GLint h_uMatShine;
GLint h_uLightMethod;
GLint h_uDirLight;
GLint h_aTexCoord;
GLint h_uTexUnit;

GLuint LightBuffObj;
GLuint LIndxBuffObj;
GLuint GIndxBuffObj;
GLuint GNBuffObj;
GLuint CubeBuffObj;
GLuint CIndxBuffObj;
GLuint NormalBuffObj;
GLuint PyramidBuffObj;
GLuint PIndxBuffObj;
GLuint PNormalBuffObj;
GLuint SkyBuffObj;
GLuint SIndxBuffObj;
GLuint STexBuffObj;

int g_LiboLen, g_GiboLen, g_CiboLen, g_PiboLen, g_SiboLen;

static const float g_groundElevation = 0;    // y coordinate of the ground
static const float g_groundWidth = 7.0; // half the ground width
static const float g_groundLength = 70.0;

RenderingHelper ModelTrans;
ObstacleList Obstacles = ObstacleList();
ParticleList Particles = ParticleList();
Image *TextureImage;
bool keyDown[256] = {0};

vec3 runnerPos = vec3(0, 0, 0);
float runnerA = 180;

float conversion = 25.0;
vec3 sHead =  vec3( 7.0,  9.0,  8.0);
vec3 sNeck =  vec3( 5.0,  4.0,  4.0);
vec3 sTorso = vec3(16.0, 20.0, 11.0);
vec3 sUArm =  vec3( 4.0, 13.0,  5.0);
vec3 sLArm =  vec3( 4.0, 11.0,  4.0);
vec3 sHand =  vec3( 2.0,  8.0,  5.0);
vec3 sWaist = vec3(16.0,  7.0, 11.0);
vec3 sULeg =  vec3( 8.0, 12.0,  8.0);
vec3 sLLeg =  vec3( 6.0, 19.0,  6.0);

vec3 head =  sHead / conversion;
vec3 neck =  sNeck / conversion;
vec3 torso = sTorso / conversion;
vec3 uArm =  sUArm / conversion;
vec3 lArm =  sLArm / conversion;
vec3 hand =  sHand / conversion;
vec3 waist = sWaist / conversion;
vec3 uLeg =  sULeg / conversion;
vec3 lLeg =  sLLeg / conversion;

float waistA = 15;
float lShoulderA = -74 + waistA; // 180 relative - waistA
float rShoulderA =  36 + waistA; //  
float lElbowA = -10 - lShoulderA; // 180Relative - shoulder
float rElbowA = 193 - rShoulderA; // 
float lHipA = 80;
float rHipA = -40;
float lKneeA = 15 + lHipA;
float rKneeA = 50 + rHipA;

float waistBackup;
float lShoulderBackup;
float rShoulderBackup;
float lHipBackup;
float rHipBackup;
float lElbowBackup;
float rElbowBackup;
float lKneeBackup;
float rKneeBackup;

bool gameOver = false;
bool drawing = true;
int score = 0;
int bestScore = 0;
int bestLevel = 1;
int level = 1;  
int levelUp = 500;

void SetMaterial(int id) {
   glUseProgram(ShadeProg);
   switch (id) {
      case 0:
         safe_glUniform3f(h_uMatDif, 0, 0.08, 0.5);
         safe_glUniform3f(h_uMatSpec, 0.4, 0.4, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.2, 0.2, 0.2);
         safe_glUniform1f(h_uMatShine, 200.0);
         break;
      case 1: // shiny blue plastic
         safe_glUniform3f(h_uMatDif, 0.0, 0.08, 0.5);
         safe_glUniform3f(h_uMatSpec, 0.14, 0.14, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.02, 0.02, 0.1);
         safe_glUniform1f(h_uMatShine, 120.0);
         break;
      case 2: // dull grey
         safe_glUniform3f(h_uMatDif, 0.3, 0.3, 0.4);
         safe_glUniform3f(h_uMatSpec, 0.3, 0.3, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.13, 0.13, 0.14);
         safe_glUniform1f(h_uMatShine, 4.0);
         break;
      case 3: // gold
         safe_glUniform3f(h_uMatDif, 0.91, 0.782, 0.82);
         safe_glUniform3f(h_uMatSpec, 1.0, 0.913, 0.8);
         safe_glUniform3f(h_uMatAmb, 0.09, 0.07, 0.08);
         safe_glUniform1f(h_uMatShine, 200.0);
         break;
      case 4:
         safe_glUniform3f(h_uMatDif, 0.6, 0.1, 0.2);
         safe_glUniform3f(h_uMatSpec, 0.6, 0.2, 0.5);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 20.0);
         break;
      case 5:
         safe_glUniform3f(h_uMatDif, 0.1, 0.8, 0.1);
         safe_glUniform3f(h_uMatSpec, 0.4, 0.5, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 1.0);
         break;
      case 6:
         safe_glUniform3f(h_uMatDif, 0.6, 0.1, 0.6);
         safe_glUniform3f(h_uMatSpec, 0.4, 0.4, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 1.5);
         break;
      case 7:
         safe_glUniform3f(h_uMatDif, 0.8, 0.3, 0.5);
         safe_glUniform3f(h_uMatSpec, 0.4, 0.2, 0.4);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 3.0);
         break;
      case 8:
         safe_glUniform3f(h_uMatDif, 0.5, 0.6, 0.5);
         safe_glUniform3f(h_uMatSpec, 0.4, 0.4, 0.8);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 1.0);
         break;
      case 9: // grass
         safe_glUniform3f(h_uMatDif, 0.3, 0.5, 0.3);
         safe_glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.1, 0.05);
         safe_glUniform1f(h_uMatShine, .5);
         break;
   }
}


void SetMaterial(vec3 diff, vec3 spec, vec3 amb, float shine) {
   glUseProgram(ShadeProg);
   safe_glUniform3f(h_uMatDif, diff.x, diff.y, diff.z);
   safe_glUniform3f(h_uMatSpec, spec.x, spec.y, spec.z);
   safe_glUniform3f(h_uMatAmb, amb.x, amb.y, amb.z);
   safe_glUniform1f(h_uMatShine, shine);
}


void SetProjectionMatrix() {
   safe_glUniformMatrix4fv(h_uProjMatrix, value_ptr(perspective(60.0f,
      (float)g_width/g_height, 0.1f, 70.f)));
}


void SetView() {
   safe_glUniformMatrix4fv(h_uViewMatrix, value_ptr(lookAt(g_camPos, g_lookPos, up)));
}


void SetModel(mat4 mat) {
   safe_glUniformMatrix4fv(h_uModelMatrix, value_ptr(mat));
}

void SetModel() {
   safe_glUniformMatrix4fv(h_uModelMatrix, value_ptr(ModelTrans.modelViewMatrix));
}

static void initSky() {
   float SkyPos[] = { 
      -55, 55, -37, 
      -55, -55, -37, 
       55, 55,  -37, 
       55, -55, -37
   };
   
   static GLfloat SkyTex[] = {
      0, 1,
      0, 0,
      1, 1,
      1, 0
   }; 

   unsigned short idx[] = {
      0, 1, 2,
      0, 2, 3
   };

   g_SiboLen = 6;
   glGenBuffers(1, &SkyBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, SkyBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(SkyPos), SkyPos, GL_STATIC_DRAW);

   glGenBuffers(1, &SIndxBuffObj);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, SIndxBuffObj);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

   glGenBuffers(1, &STexBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, STexBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(SkyTex), SkyTex, GL_STATIC_DRAW);
}

static void initGround() {
   float GrndPos[] = { 
      -g_groundWidth, g_groundElevation, -g_groundLength, 
      -g_groundWidth, g_groundElevation,  10, 
       g_groundWidth, g_groundElevation,  10, 
       g_groundWidth, g_groundElevation, -g_groundLength
   };
   
   float GrndNorm[] = { 
      0, .8, .2, 
      0, 1, 0, 
      0, 1, 0, 
      0, .8, .2
   };

   unsigned short idx[] = {
      0, 1, 2,
      0, 2, 3
   };

   g_GiboLen = 6;
   glGenBuffers(1, &GrndBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GrndPos), GrndPos, GL_STATIC_DRAW);

   glGenBuffers(1, &GIndxBuffObj);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

   glGenBuffers(1, &GNBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, GNBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GrndNorm), GrndNorm, GL_STATIC_DRAW);
}


static void initCube() {

   float CubePos[] = {
      -0.5, -0.5, -0.5, // bottom face 4 verts: 0
       0.5, -0.5, -0.5, //1
      -0.5, -0.5,  0.5, //2
       0.5, -0.5,  0.5, //3
      -0.5,  0.5, -0.5, // top face 4 verts: 4
       0.5,  0.5, -0.5, //5
      -0.5,  0.5,  0.5, //6
       0.5,  0.5,  0.5, //7
      -0.5, -0.5, -0.5, // back face 4 verts: 8 
       0.5, -0.5, -0.5, //9
      -0.5,  0.5, -0.5, //10
       0.5,  0.5, -0.5, //11
      -0.5, -0.5,  0.5, // front face 4 verts: 12
       0.5, -0.5,  0.5, //13
      -0.5,  0.5,  0.5, //14
       0.5,  0.5,  0.5, //15
      -0.5, -0.5, -0.5, // left face 4 verts: 16
      -0.5, -0.5,  0.5, //17
      -0.5,  0.5, -0.5, //18
      -0.5,  0.5,  0.5, //19
       0.5, -0.5, -0.5, // right face 4 verts: 20
       0.5, -0.5,  0.5, //21
       0.5,  0.5, -0.5, //22
       0.5,  0.5,  0.5  //23
   };

   unsigned short idx[] = {
      0, 1, 2,  // bottom face
      1, 2, 3,
      4, 5, 6,  // top face
      5, 6, 7,
      8, 9, 10,  // back face
      9, 10, 11,  
      12, 13, 14, // front face
      13, 14, 15,  
      16, 17, 18, // left face
      17, 18, 19,
      20, 21, 22, // right face
      21, 22, 23
   };
 
   float CubeNormal[] = {
      0, -1, 0, // bottom face 4 verts: 0
      0, -1, 0,
      0, -1, 0,
      0, -1, 0,
      0, 1, 0,  // top face 4 verts: 4
      0, 1, 0,
      0, 1, 0,
      0, 1, 0,
      0, 0, -1, // back face 4 verts: 8 
      0, 0, -1,  
      0, 0, -1,  
      0, 0, -1,  
      0, 0, 1,  // front face 4 verts: 12
      0, 0, 1,  
      0, 0, 1,  
      0, 0, 1,  
      -1, 0, 0, // left face 4 verts: 16
      -1, 0, 0,
      -1, 0, 0,
      -1, 0, 0,
      1, 0, 0,  // right face 4 verts: 20
      1, 0, 0,
      1, 0, 0,
      1, 0, 0  
   }; 

   g_CiboLen = 36;
   glGenBuffers(1, &CubeBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, CubeBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(CubePos), CubePos, GL_STATIC_DRAW);

   glGenBuffers(1, &CIndxBuffObj);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, CIndxBuffObj);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

   glGenBuffers(1, &NormalBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, NormalBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(CubeNormal), CubeNormal, GL_STATIC_DRAW);
}


static void initPyramid() {
   float PyramidPos[] = {
      -0.5, -0.5, -0.5, // bottom face 4 verts: 0
       0.5, -0.5, -0.5, //1
      -0.5, -0.5,  0.5, //2
       0.5, -0.5,  0.5, //3
      -0.5, -0.5, -0.5, // back face 3 verts: 4 
       0.5, -0.5, -0.5, //5
       0.0,  0.5,  0.0, //6
      -0.5, -0.5,  0.5, // front face 3 verts: 7
       0.5, -0.5,  0.5, //8
       0.0,  0.5,  0.0, //9
      -0.5, -0.5, -0.5, // left face 3 verts: 10
      -0.5, -0.5,  0.5, //11
       0.0,  0.5,  0.0, //12
       0.5, -0.5, -0.5, // right face 3 verts: 13
       0.5, -0.5,  0.5, //14
       0.0,  0.5,  0.0  //15
   };

   unsigned short idx[] = {
      0, 1, 2,  // bottom face
      1, 2, 3,
      4, 5, 6,  // back face
      7, 8, 9, // front face
      10, 11, 12, // left face
      13, 14, 15, // right face
   };
 
   float PyramidNormal[] = {
       0, -1, 0, // bottom face 4 verts: 0
       0, -1, 0, //1
       0, -1, 0, //2
       0, -1, 0, //3
       0, 0.894427, -0.447214, // back face 3 verts: 4 
       0, 0.894427, -0.447214, //5
       0, 0.894427, -0.447214, //6
       0, 0.894427, 0.447214, // front face 3 verts: 7
       0, 0.894427, 0.447214, //8
       0, 0.894427, 0.447214, //9
      -0.447214, 0.894427, 0, // left face 3 verts: 10
      -0.447214, 0.894427, 0, //11
      -0.447214, 0.894427, 0, //12
       0.447214, 0.894427, 0, // right face 3 verts: 13
       0.447214, 0.894427, 0, //14
       0.447214, 0.894427, 0  //15
   }; 

   g_PiboLen = 18;
   glGenBuffers(1, &PyramidBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, PyramidBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(PyramidPos), PyramidPos, GL_STATIC_DRAW);

   glGenBuffers(1, &PIndxBuffObj);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, PIndxBuffObj);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

   glGenBuffers(1, &PNormalBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, PNormalBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(PyramidNormal), PyramidNormal, GL_STATIC_DRAW);
}


void InitGeom() {
   initSky();
   initGround();
   initCube();
   initPyramid();
}


int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
   GLuint VS, FS; //handles to shader object
   GLint vCompiled, fCompiled, linked; //status of shader
   
   VS = glCreateShader(GL_VERTEX_SHADER);
   FS = glCreateShader(GL_FRAGMENT_SHADER);
   
   glShaderSource(VS, 1, &vShaderName, NULL);
   glShaderSource(FS, 1, &fShaderName, NULL);
   
   glCompileShader(VS);
   printOpenGLError();
   glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
   printShaderInfoLog(VS);

   glCompileShader(FS);
   printOpenGLError();
   glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
   printShaderInfoLog(FS);

   if (!vCompiled || !fCompiled) {   
      printf("Error compiling the shader %s", vShaderName);
      return 0;
   }
    
   ShadeProg = glCreateProgram();
   glAttachShader(ShadeProg, VS);
   glAttachShader(ShadeProg, FS);
   
   glLinkProgram(ShadeProg);
   printOpenGLError();
   glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
   printProgramInfoLog(ShadeProg);

   glUseProgram(ShadeProg);

   h_aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
   h_aNormal = safe_glGetAttribLocation(ShadeProg,   "aNormal");
   h_uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
   h_uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
   h_uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
   h_viewPos = safe_glGetUniformLocation(ShadeProg, "viewPos");
   h_uDirLight = safe_glGetUniformLocation(ShadeProg, "dirLight");
   h_uLightMethod = safe_glGetUniformLocation(ShadeProg, "uLightMethod");
   h_uMatDif = safe_glGetUniformLocation(ShadeProg, "uMat.dColor");
   h_uMatSpec = safe_glGetUniformLocation(ShadeProg, "uMat.sColor");
   h_uMatAmb = safe_glGetUniformLocation(ShadeProg, "uMat.aColor");
   h_uMatShine = safe_glGetUniformLocation(ShadeProg, "uMat.shine");
   h_aTexCoord = safe_glGetAttribLocation(ShadeProg,  "aTexCoord");
   h_uTexUnit = safe_glGetUniformLocation(ShadeProg, "uTexUnit");

   printf("sucessfully installed shader %d\n", ShadeProg);
   return 1;
}


void drawCube(int matId) {
   SetModel();
   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, CubeBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, CIndxBuffObj);
   SetMaterial(matId);  
   safe_glEnableVertexAttribArray(h_aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, NormalBuffObj);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glDrawElements(GL_TRIANGLES, g_CiboLen, GL_UNSIGNED_SHORT, 0);
   safe_glDisableVertexAttribArray(h_aPosition);
}


void drawCube(vec3 d, vec3 s, vec3 a, float shine) {
   SetModel();
   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, CubeBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, CIndxBuffObj);
   SetMaterial(d, s, a, shine);
   safe_glEnableVertexAttribArray(h_aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, NormalBuffObj);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glDrawElements(GL_TRIANGLES, g_CiboLen, GL_UNSIGNED_SHORT, 0);
   safe_glDisableVertexAttribArray(h_aPosition);
}


void drawPyramid(int matId) {
   SetModel();
   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, PyramidBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, PIndxBuffObj);
   SetMaterial(matId);  
   safe_glEnableVertexAttribArray(h_aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, PNormalBuffObj);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glDrawElements(GL_TRIANGLES, g_PiboLen, GL_UNSIGNED_SHORT, 0);
   safe_glDisableVertexAttribArray(h_aPosition);
}


void drawParticle(float x, float y, float z, vec3 diff, vec3 spec, vec3 amb, float shine) {
   float width = .4, height = 0.4;

   y += height / 2.0 + 0.01;

   ModelTrans.loadIdentity();
   SetModel();

   ModelTrans.pushMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(vec3(x, y, z));
         ModelTrans.scale(width, width, width);
         drawCube(diff, spec, amb, shine);
      ModelTrans.popMatrix();
   ModelTrans.popMatrix();
}


void drawPole(float x, float y, float z) {
   float width = 0.5, height = 4.0;

   y += height / 2.0 + 0.01;

   ModelTrans.loadIdentity();
   SetModel();

   ModelTrans.pushMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(vec3(x, y , z));
         ModelTrans.scale(width, height, width);
         drawCube(1);
      ModelTrans.popMatrix();
   ModelTrans.popMatrix();
}


void drawWall(float x, float y, float z, float width) {
   float height = 3.5, length = 0.5;

   y += height / 2.0 + 0.01;

   ModelTrans.loadIdentity();
   SetModel();

   ModelTrans.pushMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(vec3(x, height / 2.0, z));
         ModelTrans.scale(width, height, length);
         drawCube(1);
      ModelTrans.popMatrix();
   ModelTrans.popMatrix();
}


void drawSpikes(float x, float y, float z, int num) {
   float initX = 0.5 + (float)num / 2.0 * -1.0;
   int i = 1;
   y += 0.5 + 0.001;

   ModelTrans.loadIdentity();
   SetModel();

   ModelTrans.pushMatrix();
      // starting on left, draw first spike
      ModelTrans.translate(vec3(initX + x, y, z));
      drawPyramid(0);
      for (; i < num; i++) {
         ModelTrans.pushMatrix();
            ModelTrans.translate(vec3(1.0, 0.0, 0.0));
            drawPyramid(0);
      }
      for (; i > 0; i--) {
         ModelTrans.popMatrix();
      }
}


void drawRunner(float x, float y, float z, float angle) {
   float edge = 0.001;
   y += waist.y / 2.0 + uLeg.y + lLeg.y;
   
   if (!drawing)
      return;
   ModelTrans.loadIdentity();
   SetModel();

   ModelTrans.pushMatrix();
      // waist is the root
      ModelTrans.translate(vec3(x, y, z));
      ModelTrans.scale(waist.x, waist.y, waist.z);
      drawCube(1);
      ModelTrans.rotate(angle, up);
      ModelTrans.pushMatrix();
         // waist joint, dont draw
         ModelTrans.scale(1.0 / waist.x, 1.0 / waist.y, 1.0 / waist.z);
         ModelTrans.translate(vec3(0, waist.y / 2.0, 0));
         ModelTrans.rotate(waistA, vec3(1, 0, 0));
         ModelTrans.pushMatrix();
            // torso
            ModelTrans.translate(vec3(0, torso.y / 2.0, 0));
            ModelTrans.scale(torso.x + edge, torso.y, torso.z - edge);
            drawCube(2);
            ModelTrans.pushMatrix(); 
               // left shoulder joint, dont draw
               ModelTrans.scale(1.0 / torso.x, 1.0 / torso.y, 1.0 / torso.z);
               ModelTrans.translate(vec3(torso.x / 2.0, torso.y / 2.0, 0));
               ModelTrans.rotate(lShoulderA, vec3(-1, 0, 0));
               ModelTrans.pushMatrix(); 
                  // left upper arm
                  ModelTrans.translate(vec3(uArm.x / 2.0, -uArm.y / 2.0, 0));
                  ModelTrans.scale(uArm.x, uArm.y, uArm.z);
                  drawCube(3);
                  ModelTrans.pushMatrix(); 
                     // left elbow joint, dont draw
                     ModelTrans.scale(1.0 / uArm.x, 1.0 / uArm.y, 1.0 / uArm.z);
                     ModelTrans.translate(vec3(0, -uArm.y / 2.0, 0));
                     ModelTrans.rotate(lElbowA, vec3(-1, 0, 0));
                     ModelTrans.pushMatrix(); 
                        // left lower arm
                        ModelTrans.translate(vec3(0, -lArm.y / 2.0, 0));
                        ModelTrans.scale(lArm.x - edge, lArm.y, lArm.z);
                        drawCube(4);
                     ModelTrans.popMatrix(); 
                  ModelTrans.popMatrix(); 
               ModelTrans.popMatrix(); 
            ModelTrans.popMatrix();
            ModelTrans.pushMatrix();
               // right shoulder joint, dont draw
               ModelTrans.scale(1.0 / torso.x, 1.0 / torso.y, 1.0 / torso.z);
               ModelTrans.translate(vec3(-torso.x / 2.0, torso.y / 2.0, 0));
               ModelTrans.rotate(rShoulderA, vec3(-1, 0, 0));
               ModelTrans.pushMatrix(); 
                  // right upper arm
                  ModelTrans.translate(vec3(-uArm.x / 2.0, -uArm.y / 2.0, 0));
                  ModelTrans.scale(uArm.x, uArm.y, uArm.z);
                  drawCube(3);
                  ModelTrans.pushMatrix(); 
                     // right elbow joint, dont draw
                     ModelTrans.scale(1.0 / uArm.x, 1.0 / uArm.y, 1.0 / uArm.z);
                     ModelTrans.translate(vec3(0, -uArm.y / 2.0, 0));
                     ModelTrans.rotate(rElbowA, vec3(-1, 0, 0));
                     ModelTrans.pushMatrix(); 
                        // right lower arm
                        ModelTrans.translate(vec3(0, -lArm.y / 2.0, 0));
                        ModelTrans.scale(lArm.x - edge, lArm.y, lArm.z);
                        drawCube(4);
                     ModelTrans.popMatrix();
                  ModelTrans.popMatrix();
               ModelTrans.popMatrix();
            ModelTrans.popMatrix();
            ModelTrans.pushMatrix();
               // rigid neck
               ModelTrans.scale(1.0 / torso.x, 1.0 / torso.y, 1.0 / torso.z);
               ModelTrans.translate(vec3(0, torso.y / 2.0 + neck.y / 2.0, 0));
               ModelTrans.scale(neck.x, neck.y, neck.z);
               drawCube(5);
               ModelTrans.pushMatrix();
                  // rigid head
                  ModelTrans.scale(1.0 / neck.x, 1.0 / neck.y, 1.0 / neck.z);
                  ModelTrans.translate(vec3(0, neck.y / 2.0 + head.y / 2.0, 0));
                  ModelTrans.scale(head.x, head.y, head.z);
                  drawCube(6);
               ModelTrans.popMatrix();
            ModelTrans.popMatrix();
         ModelTrans.popMatrix();
      ModelTrans.popMatrix();
      ModelTrans.pushMatrix();
         // left hip joint, dont draw
         ModelTrans.scale(1.0 / waist.x, 1.0 / waist.y, 1.0 / waist.z);
         ModelTrans.translate(vec3(waist.x / 2.0 - uLeg.x / 2.0, -waist.y / 2.0, 0));
         ModelTrans.rotate(lHipA, vec3(-1, 0, 0));
         ModelTrans.pushMatrix();
            // left upper leg
            ModelTrans.translate(vec3(0, -uLeg.y / 2.0, 0));
            ModelTrans.scale(uLeg.x - edge, uLeg.y, uLeg.z);
            drawCube(3);
            ModelTrans.pushMatrix();
               // left kneww joint, dont draw
               ModelTrans.scale(1.0 / uLeg.x, 1.0 / uLeg.y, 1.0 / uLeg.z);
               ModelTrans.translate(vec3(0, -uLeg.y / 2.0, 0));
               ModelTrans.rotate(lKneeA, vec3(1, 0, 0));
               ModelTrans.pushMatrix();
                  // left lower leg
                  ModelTrans.translate(vec3(0, -lLeg.y / 2.0, 0));
                  ModelTrans.scale(lLeg.x, lLeg.y, lLeg.z);
                  drawCube(4);
               ModelTrans.popMatrix();
            ModelTrans.popMatrix();
         ModelTrans.popMatrix();
      ModelTrans.popMatrix();
      ModelTrans.pushMatrix();
         // right hip joint, dont draw
         ModelTrans.scale(1.0 / waist.x, 1.0 / waist.y, 1.0 / waist.z);
         ModelTrans.translate(vec3(-waist.x / 2.0 + uLeg.x / 2.0, -waist.y / 2.0, 0));
         ModelTrans.rotate(rHipA, vec3(-1, 0, 0));
         ModelTrans.pushMatrix();
            // right upper leg
            ModelTrans.translate(vec3(0, -uLeg.y / 2.0, 0));
            ModelTrans.scale(uLeg.x - edge, uLeg.y, uLeg.z);
            drawCube(3);
            ModelTrans.pushMatrix();
               // right knee joint, dont draw
               ModelTrans.scale(1.0 / uLeg.x, 1.0 / uLeg.y, 1.0 / uLeg.z);
               ModelTrans.translate(vec3(0, -uLeg.y / 2.0, 0));
               ModelTrans.rotate(rKneeA, vec3(1, 0, 0));
               ModelTrans.pushMatrix();
                  // right lower leg
                  ModelTrans.translate(vec3(0, -lLeg.y / 2.0, 0));
                  ModelTrans.scale(lLeg.x, lLeg.y, lLeg.z);
                  drawCube(4);
               ModelTrans.popMatrix();
            ModelTrans.popMatrix();
         ModelTrans.popMatrix();
      ModelTrans.popMatrix();
   ModelTrans.popMatrix();
}


void Initialize () {
   glClearColor (0.15f, 0.15f, 0.7f, 1.0f);                        
   glEnable (GL_DEPTH_TEST);// Enable Depth Testing
   ModelTrans.useModelViewMatrix();
   ModelTrans.loadIdentity();
   glEnable(GL_TEXTURE_2D);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}


int genObstacle(int last, int level) {
   int type = rand() % 3 + 1;
   int poleTypes = 3;
   float x;
   float z = -80;
   float step = 0.2;
   float maxSize = 2 * g_groundWidth;
   int num;
   float width;
   float percent = (float)rand() / RAND_MAX; // 0 <= result <= 1.0
   float xRand = percent * 2 - 1; // -1 <= result <= 1

   int config;
   float squeeze;
   int direction;
   int rows = 18;
   float dx = 0.4; // change in x
   float dz = 0.8; // change in z
   float space = 3.5; // bodywidths between poles
   float minSpace = 2;
   float multiplier = 1;
   float multiple = 1.1;

   while (level-- > 1) {
      multiplier *= multiple;
      space *= 0.90;

      if (space < minSpace) {
         space = minSpace;
      }
   }

   rows *= multiplier;
   if (last == POLE) {
      type = rand() % 2 + 2;
   }
   //printf("sign:%f\n", xRand);
   switch (type) {
      case POLE:
         config = rand() % poleTypes;
         squeeze = waist.x + 2 * uArm.x;
         squeeze *= space;
         
         switch (config) {
            case 0: // guided path
               direction = rand() % 2 ? 1 : -1;
               x = xRand * (g_groundWidth - squeeze) - xRand * 0.25; // Center of Poles
               
               for (float i = x - squeeze; i > -g_groundWidth - 0.25; i -= 0.75)
                  Obstacles.add(type, i, z, step, 0);
               for (float i = x + squeeze; i < g_groundWidth + 0.25; i += 0.75)
                  Obstacles.add(type, i, z, step, 0);
               for (int i = 0; i < rows; i++) {
                  if (x - squeeze + direction * dx < -g_groundWidth || x + squeeze + direction * dx > g_groundWidth) {
                     direction *= -1;
                  } else {
                     x += direction * dx;
                  }
                  Obstacles.add(type, x - squeeze, z, step, 0);
                  Obstacles.add(type, x + squeeze, z, step, 0);
                  z -= dz;
               }
               for (float i = x - squeeze; i > -g_groundWidth - 0.25; i -= 0.75)
                  Obstacles.add(type, i, z + dz, step, 0);
               for (float i = x + squeeze; i < g_groundWidth + 0.25; i += 0.75)
                  Obstacles.add(type, i, z + dz, step, 0);

               break;
            case 1: // random "garden"
               rows /= 5;

               for (int i = 0; i < rows; i++) {
                  percent = (float)rand() / RAND_MAX;
                  x = percent * squeeze - g_groundWidth + 0.25;
               
                  while (x < g_groundWidth - 0.25) {
                     Obstacles.add(type, x, z - (percent * 3 - 1.5) * squeeze, step, 0);
                     x += (1 + percent) * squeeze;
                  }
                  z -= 5;
               }
               break;
            case 2: // zigzag path
               direction = rand() % 2 ? 1 : -1;
               x = xRand * (g_groundWidth - squeeze) - xRand * 0.25; // Center of Poles
               
               for (float i = x - squeeze; i > -g_groundWidth - 0.25; i -= 0.75)
                  Obstacles.add(type, i, z, step, 0);
               for (float i = x + squeeze; i < g_groundWidth + 0.25; i += 0.75)
                  Obstacles.add(type, i, z, step, 0);
               for (int i = 0; i < rows; i++) {
                  if (i % 3 == 0) {
                     printf("Dat zigzag doe\n");
                     direction = rand() % 2 ? 1 : -1;
                  }
                  //printf("i\% 3: %d direction: %d\n", i%3, direction);
                     
                  if (x - squeeze + direction * dx < -g_groundWidth || x + squeeze + direction * dx > g_groundWidth) {
                     direction *= -1;
                  } else {
                     x += direction * dx;
                  }
                  Obstacles.add(type, x - squeeze, z, step, 0);
                  Obstacles.add(type, x + squeeze, z, step, 0);
                  z -= dz;
               }
               for (float i = x - squeeze; i > -g_groundWidth - 0.25; i -= 0.75)
                  Obstacles.add(type, i, z + dz, step, 0);
               for (float i = x + squeeze; i < g_groundWidth + 0.25; i += 0.75)
                  Obstacles.add(type, i, z + dz, step, 0);

               break;
         }
         //cout << "Adding Pole at " << x << endl;
         break;
      case WALL:
         width = maxSize;//percent * (maxSize - 1) + 1;
         x = 0;//(maxSize - width) / 2 * xRand;
         Obstacles.add(type, x, z, step, width);

         //cout << "Adding Wall at " << x << " width: " << width << endl;
         break;
      case SPIKE:
         num = (int)maxSize;//(int)(percent * maxSize + 1);
         x = 0;//(maxSize - num) / 2 * xRand;
         Obstacles.add(type, x, z, step, num);
         //cout << "Adding " << num << " Spikes at " << x << endl;
         break;
   }

   return type;
}


void drawObstacles() {
   if (!Obstacles.size)
      return;

   Obstacles.draw(drawPole, drawWall, drawSpikes);
}


void drawParticles() {
   if (!Particles.size)
      return;

   Particles.draw(drawParticle);
}


void reset() {
   if (gameOver) {
      while (Obstacles.size) {
         Obstacles.remove();
      }
      while (Particles.size) {
         Particles.remove();
      }
      runnerPos = vec3(0, 0, 0);
      score = 0;
      level = 1;
      levelUp = 1000;
      gameOver = false;
      drawing = true;
      for (int i = 0; i < 256; i++)
         keyDown[i] = false;

      g_camPos = fixedCam;
      alpha = 2 * pi / 3, beta = -pi / 2.0;
      g_lookPos = vec3(g_camPos.x + cos(beta) * sin(alpha),
                      g_camPos.y + cos(alpha),
                      g_camPos.z + sin(beta) * sin(alpha));
      g_gazeVec = normalize(g_lookPos - g_camPos);
      g_vVec, g_wVec = -g_gazeVec;
      g_uVec = normalize(cross(up, g_wVec));


      glutIdleFunc(tick);
   }
}


void Draw(void) {
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glUseProgram(ShadeProg);
   
   SetProjectionMatrix();
   SetView();

   safe_glUniform3f(h_uDirLight, dirLight.x, dirLight.y, dirLight.z);
   glUniform3f(h_viewPos, g_camPos.x, g_camPos.y, g_camPos.z);

   // Set up the texture unit
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, 0);

   // Use texture 0 (only texture)
   safe_glUniform1i(h_uTexUnit, 0);

   // Draw Sky "wall"
   safe_glUniform1i(h_uLightMethod, 2);
   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, SkyBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   SetModel(mat4());
   safe_glEnableVertexAttribArray(h_aTexCoord);
   glBindBuffer(GL_ARRAY_BUFFER, STexBuffObj);
   safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 
   
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, SIndxBuffObj);
   glDrawElements(GL_TRIANGLES, g_SiboLen, GL_UNSIGNED_SHORT, 0);


   // Draw ground
   safe_glUniform1i(h_uLightMethod, lightMethod);
   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
   SetModel(mat4());
   SetMaterial((materialID + 8) % 10);
   safe_glEnableVertexAttribArray(h_aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, GNBuffObj);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);

   drawRunner(runnerPos.x, runnerPos.y, runnerPos.z, runnerA);
   drawObstacles();
   drawParticles();

   //clean up 
   safe_glDisableVertexAttribArray(h_aPosition);
   safe_glDisableVertexAttribArray(h_aNormal);
   safe_glDisableVertexAttribArray(h_aTexCoord);
   glUseProgram(0);   
   glDisable(GL_TEXTURE_2D);
   glutSwapBuffers();
}


void ReshapeGL (int width, int height) {
   g_width = (float)width;
   g_height = (float)height;
   glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));
}


void Keyboard(unsigned char key, int x, int y) {
   float step = 0.7;

   switch (key) {
      case 's':
         g_camPos += step * g_wVec;
         g_lookPos += step * g_wVec;
         break;
      case 'w':
         g_camPos -= step * g_wVec;
         g_lookPos -= step * g_wVec;
         break;
      case 'd': case 'D':
         keyDown['D'] = true;
         break;
      case 'a': case 'A':
         keyDown['A'] = true;
         break;
      case ' ':
         keyDown[' '] = true;
         break;
      case 'v': case 'V':
         keyDown['V'] = true;
         break;
      case 'r':
         reset();
         break;
      case 'n': // Draw normals
         //lightMethod = 0;
         glutIdleFunc(NULL);
         break;
      case 'N':
         glutIdleFunc(tick);
         break;
      case 'p': // Phong shading
         lightMethod = 1;
         break;
      case 'q': case 'Q' :
         exit(EXIT_SUCCESS);
      break;
   }
   glutPostRedisplay();
}


void KeyboardUp(unsigned char key, int x, int y) {
   switch (key) {
      case 'd':
         keyDown['D'] = false;
         runnerA = 180;
         break;
      case 'a':
         keyDown['A'] = false;
         runnerA = 180;
         break;
      case ' ':
         keyDown[' '] = false;
         break;
      case 'v':
         keyDown['V'] = false;
         break;
   }

   glutPostRedisplay();
}


void SpecialKey(int key, int x, int y) {
   switch (key) {
      case GLUT_KEY_LEFT:
         keyDown[GLUT_KEY_LEFT] = true;
         break;
      case GLUT_KEY_RIGHT:
         keyDown[GLUT_KEY_RIGHT] = true;
         break;
      case GLUT_KEY_UP:
         keyDown[GLUT_KEY_UP] = true;
         break;
   }

   glutPostRedisplay();
}


void SpecialKeyUp(int key, int x, int y) {
   switch (key) {
      case GLUT_KEY_LEFT:
         keyDown[GLUT_KEY_LEFT] = false;
         runnerA = 180;
         break;
      case GLUT_KEY_RIGHT:
         keyDown[GLUT_KEY_RIGHT] = false;
         runnerA = 180;
         break;
      case GLUT_KEY_UP:
         keyDown[GLUT_KEY_UP] = false;
         break;
   }

   glutPostRedisplay();
}


void saveRunner() {
   waistBackup = waistA;
   lShoulderBackup = lShoulderA;
   rShoulderBackup = rShoulderA;
   lHipBackup = lHipA;
   rHipBackup = rHipA;
   lElbowBackup = lElbowA;
   rElbowBackup = rElbowA;
   lKneeBackup = lKneeA;
   rKneeBackup = rKneeA;
}


void restoreRunner() {
   waistA = waistBackup;
   lShoulderA = lShoulderBackup;
   rShoulderA = rShoulderBackup;
   lHipA = lHipBackup;
   rHipA = rHipBackup;
   lElbowA = lElbowBackup;
   rElbowA = rElbowBackup;
   lKneeA = lKneeBackup;
   rKneeA = rKneeBackup;
}


void running() {
   float shoulderRange = 110;
   float hipRange = 120;
   float elbowRange = 94;
   float kneeRange = 85;
   
   static int frames = 20;
   static int frame = 0;
   static float dShoulder = shoulderRange / frames;
   static float dHip = hipRange / frames;
   static float dElbow = elbowRange / frames;
   static float dKnee = kneeRange / frames;

   if (frame == frames) {
      frame = 0;
      dHip *= -1;
      dShoulder *= -1;
      dElbow *= -1;
      dKnee *= -1;
   } else {
      lShoulderA += dShoulder;
      rShoulderA -= dShoulder;
      lHipA -= dHip;
      rHipA += dHip;

      if ((float)frame / (float)frames < 0.5) {
         lKneeA -= 2 * dKnee;
         rKneeA += 2 * dKnee;
      }
      lElbowA += dElbow;
      rElbowA -= dElbow;
      frame++;
   }
}


void jumping() {
   lShoulderA = -100;
   rShoulderA = 77;
   lElbowA = 80;
   rElbowA = 90;
   lHipA = 120;
   rHipA = -70;
   lKneeA = 90;
   rKneeA = 50;
}


void punching(int tick, int ticks) {
   float up = 0.05;
   float percent = (float)tick / (float)ticks;

   waistA = 55;
   lShoulderA = -55;
   rShoulderA = 130;
   lElbowA = 115;
   rElbowA = 10;
   lHipA = 70;
   rHipA = -45;
   lKneeA = 100;
   rKneeA = 7;
   
   if (percent > .8) {
      runnerPos.y += 2 * up;
   } else if (percent > .6) {
      runnerPos.y += up;
   } else if (percent > .4) {

   } else if (percent > .2) {
      runnerPos.y -= up;
   } else {
      runnerPos.y -= 2 * up;
   }
   if (tick == 1) 
      runnerPos.y = 0;
}


void moveRunner() {
   static int ticks = -1;
   static bool isJumping = false;
   static bool isPunching = false;
   static int reverse = 0;
   float step = 0.1;
   float jump = 0.3;
   int jumpFrames = 15;
   int punchFrames = 15;
   vec3 diff, spec, amb;
   float shine;

   // movement commands
   if ((keyDown['A'] || keyDown[GLUT_KEY_LEFT]) /*&& runnerPos.x > -g_groundWidth*/) {
      runnerPos.x -= step;
      runnerA = 200;
   } else if ((keyDown['D'] || keyDown[GLUT_KEY_RIGHT]) /*&& runnerPos.x < g_groundWidth*/) {
      runnerPos.x += step;
      runnerA = 160;
   }
   if ((runnerPos.x < -g_groundWidth - 0.25 || runnerPos.x > g_groundWidth + 0.25) && runnerPos.y <= 0) {
      glutKeyboardFunc(NULL);
      glutKeyboardUpFunc(NULL);
      glutSpecialFunc(NULL);
      glutSpecialUpFunc(NULL);
      gameOver = true;
      runnerPos.y -= step;
      if (runnerPos.y < -5) {
         glutKeyboardFunc(Keyboard);
         glutKeyboardUpFunc(KeyboardUp);
         glutSpecialFunc(SpecialKey);
         glutSpecialUpFunc(SpecialKeyUp);
         glutIdleFunc(NULL);
      }
   }
   // activate jumping/punching
   if (keyDown[' '] && !isJumping && !isPunching) {
      isJumping  = true;
      saveRunner();
   } else if ((keyDown['V'] || keyDown[GLUT_KEY_UP]) && !isPunching && !isJumping) {
      isPunching = true;
      saveRunner();
   }

   // jump up, jump up and get down !
   if (isJumping && ticks == -1) {
      ticks = jumpFrames;
      reverse = 0;
      runnerPos.y += jump * ((float)ticks / (float)jumpFrames);
      ticks--;
   } else if (isJumping) {
      if (ticks > 0) {
         runnerPos.y += jump * ((float)ticks / (float)jumpFrames);
         ticks--;
      } else {
         runnerPos.y -= jump * ((float)reverse / (float)jumpFrames);
         reverse++;
         if (reverse > jumpFrames) {
            isJumping = false;
            ticks = -1;
            restoreRunner();
         }
      }
   } else if (isPunching && ticks == -1) { // It's clobbering time!
      ticks = punchFrames;
   }  else if (isPunching) {
      if (ticks > 0) {
         ticks--;
      } else {
         isPunching = false;
         ticks = -1;
         restoreRunner();
      }
   }

   // ANIMATIONS
   if (isJumping) {
      jumping();
   } else if (isPunching) {
      punching(ticks, punchFrames);
   } else {
      running();
   }

   if (Obstacles.isColliding(runnerPos.x, runnerPos.y + lLeg.y / 2, torso.x / 2 + uArm.x, isPunching ? uArm.y + lArm.y : waist.z / 2, isPunching, &Particles)) {
      if (!gameOver) {
         drawing = false;
         for (int i = 0; i < 25; i++) {
            diff = vec3((float)rand() / (float)RAND_MAX * 0.3 + 0.7,
                        (float)rand() / (float)RAND_MAX * 0.3,
                        (float)rand() / (float)RAND_MAX * 0.3);

            spec = vec3((float)rand() / (float)RAND_MAX * 0.3,
                        (float)rand() / (float)RAND_MAX * 0.1,
                        (float)rand() / (float)RAND_MAX * 0.1);

            amb = vec3(0.2, 0.1, 0.1);
            shine = (float)rand() / (float)RAND_MAX * 0.3;
            Particles.add(vec3(runnerPos.x + randRange(.25), runnerPos.y + rand1() * .25, randRange(0.25)),
                          vec3(randRange(0.25), rand1() * 0.4 + .3, randRange(0.1)),
                          0.02, diff, spec, amb, shine);
         }
      }

      if (!Particles.size)
         glutIdleFunc(NULL);
      gameOver = true;
   }
   //printf("isPunching:%d ticks: %d\n", isPunching, ticks);
   //printf("Runner position: x=%.3f y=%.3f\r", runnerPos.x, runnerPos.y);
}


void Motion(int x, int y) {
   int dx = x - g_mouseX;
   int dy;
   float step = 1 / 57.2957795 / 2;

   y = g_height - y;
   dy =  y - g_mouseY;

   g_mouseX = x;
   g_mouseY = y;

   alpha -= dy * step;
   beta += dx * step;
   
   //cout << "pitch: " << alpha * 57.2957795 << " yaw: " << beta * 57.2957795 << endl;

   if (alpha >= pi)
      alpha = pi - step;
   else if (alpha <= 0)
      alpha = step;

   g_lookPos.x = g_camPos.x + cos(beta) * sin(alpha);
   g_lookPos.y = g_camPos.y + cos(alpha);
   g_lookPos.z = g_camPos.z + sin(beta) * sin(alpha);

   g_wVec = g_gazeVec = normalize(g_lookPos - g_camPos);
   g_wVec *= -1.0;
   g_uVec = normalize(cross(up, g_wVec));
   //printf("%.2f %.2lf %.2lf\n", g_lookPos.x, g_lookPos.y, g_lookPos.z);

   glutPostRedisplay();
}


void PasMotion(int x, int y) {
   g_mouseX = x;
   g_mouseY = g_height - y;
   glutPostRedisplay();
}


void tick() {
   int initPoleCounter = 100;
   int initCounter = 55;
   static int poleCounter = initPoleCounter;
   static int counter = initCounter;
   static int count = 0;
   static int poleCount = 0;
   static int last = 0;
   int minCount = 30;
   char buf[256] = {};

   if (count > counter && poleCount > poleCounter) {
      count = 0;
      last = genObstacle(last, level);

      if (last == POLE) {
         poleCount = 0;
      }
      if (poleCount < poleCounter) {
         last = POLE;
         poleCount++;
      }
   }
   else {
      count++;
      poleCount++;
   }

   moveRunner();
   Obstacles.move();
   Particles.move();
   if (gameOver) {
      poleCounter = initPoleCounter;
      counter = initCounter;
   }
   if (!gameOver) {
      score += level;
      
      if (score > levelUp) {
         level++;
         levelUp *= 2;
         if (counter > minCount) {
            counter--;
         }
         poleCounter *= 1.1;
         if (level > bestLevel) {
            bestLevel = level;
         }
      }
      if (score > bestScore) {
         bestScore = score;
      }
      sprintf(buf, "Level: %d  Score: %d  High Level: %d  High Score: %d", level, score, bestLevel, bestScore);
      glutSetWindowTitle(buf);
   }
   //printf("poleCount : %d  poleCounter: %d  count: %d  counter: %d  levelUp: %d\n", poleCount, poleCounter, count, counter, levelUp);

   glutPostRedisplay();
}


//routines to load in a bmp files - must be 2^nx2^m and a 24bit bmp
GLvoid LoadTexture(char* image_file, int texID) { 
  
  TextureImage = (Image *) malloc(sizeof(Image));
  if (TextureImage == NULL) {
    printf("Error allocating space for image");
    exit(1);
  }
  cout << "trying to load " << image_file << endl;
  if (!ImageLoad(image_file, TextureImage)) {
    exit(1);
  }  
  /*  2d texture, level of detail 0 (normal), 3 components (red, green, blue),            */
  /*  x size from image, y size from image,                                              */    
  /*  border 0 (normal), rgb color data, unsigned byte data, data  */ 
  glBindTexture(GL_TEXTURE_2D, texID);
  glTexImage2D(GL_TEXTURE_2D, 0, 3,
    TextureImage->sizeX, TextureImage->sizeY,
    0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage->data);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); /*  cheap scaling when image bigger than texture */    
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); /*  cheap scaling when image smalled than texture*/
  
}


/* BMP file loader loads a 24-bit bmp file only */
/*
* getint and getshort are help functions to load the bitmap byte by byte
*/
static unsigned int getint(FILE *fp) {
   int c, c1, c2, c3;

   /*  get 4 bytes */ 
   c = getc(fp);  
   c1 = getc(fp);  
   c2 = getc(fp);  
   c3 = getc(fp);

   return ((unsigned int) c) +   
    (((unsigned int) c1) << 8) + 
    (((unsigned int) c2) << 16) +
    (((unsigned int) c3) << 24);
}

static unsigned int getshort(FILE *fp){
   int c, c1;

   /* get 2 bytes*/
   c = getc(fp);  
   c1 = getc(fp);

   return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

/*  quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  */
int ImageLoad(char *filename, Image *image) {
   FILE *file;
   unsigned long size;                 /*  size of the image in bytes. */
   unsigned long i;                    /*  standard counter. */
   unsigned short int planes;          /*  number of planes in image (must be 1)  */
   unsigned short int bpp;             /*  number of bits per pixel (must be 24) */
   char temp;                          /*  used to convert bgr to rgb color. */

   /*  make sure the file is there. */
   if ((file = fopen(filename, "rb"))==NULL) {
      printf("File Not Found : %s\n",filename);
      return 0;
   }

   /*  seek through the bmp header, up to the width height: */
   fseek(file, 18, SEEK_CUR);

   /*  No 100% errorchecking anymore!!! */

   /*  read the width */    image->sizeX = getint (file);

   /*  read the height */ 
   image->sizeY = getint (file);

   /*  calculate the size (assuming 24 bits or 3 bytes per pixel). */
   size = image->sizeX * image->sizeY * 3;

   /*  read the planes */    
   planes = getshort(file);
   if (planes != 1) {
      printf("Planes from %s is not 1: %u\n", filename, planes);
      return 0;
   }

   /*  read the bpp */    
   bpp = getshort(file);
   if (bpp != 24) {
      printf("Bpp from %s is not 24: %u\n", filename, bpp);
      return 0;
   }

   /*  seek past the rest of the bitmap header. */
   fseek(file, 24, SEEK_CUR);

   /*  read the data.  */
   image->data = (char *) malloc(size);
   if (image->data == NULL) {
      printf("Error allocating memory for color-corrected image data");
      return 0; 
   }

   if ((i = fread(image->data, size, 1, file)) != 1) {
      printf("Error reading image data from %s.\n", filename);
      return 0;
   }

   for (i=0;i<size;i+=3) { /*  reverse all of the colors. (bgr -> rgb) */
      temp = image->data[i];
      image->data[i] = image->data[i+2];
      image->data[i+2] = temp;
   }

   fclose(file); /* Close the file and release the filedes */

   /*  we're done. */
   return 1;
}


int main(int argc, char *argv[]) {
   glutInit(&argc, argv);
   glutInitWindowPosition(200, 200);
   glutInitWindowSize(800, 600);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("Program4");
   glutReshapeFunc(ReshapeGL);
   glutDisplayFunc(Draw);
   glutKeyboardFunc(Keyboard);
   glutKeyboardUpFunc(KeyboardUp);
   glutSpecialFunc(SpecialKey);
   glutSpecialUpFunc(SpecialKeyUp);
   glutMotionFunc(Motion);
   glutPassiveMotionFunc(PasMotion);
   glutIdleFunc(tick);
   glutIgnoreKeyRepeat(1);
   
   g_width = g_height = 200;
   LoadTexture((char *)"sky.bmp", 0);
  
   #ifdef _WIN32 
   GLenum err = glewInit();
   if (GLEW_OK != err) {
      std::cerr << "Error initializing glew! " << glewGetErrorString(err) << std::endl;
      return 1;
   }
   #endif

   Initialize();

   getGLversion();
   if (!InstallShader(textFileRead((char *)"mesh_vert.glsl"),
    textFileRead((char *)"mesh_frag.glsl"))) {
      printf("Error installing shader!\n");
      return 0;
   }

   InitGeom();
   glutMainLoop();
   return 0;
}
