struct Material {
   vec3 dColor;
   vec3 sColor;
   vec3 aColor;
   float shine;
};

uniform mat4 uProjMatrix;
uniform mat4 uViewMatrix;
uniform mat4 uModelMatrix;

uniform sampler2D uTexUnit;

uniform int uLightMethod; // 0 for normal, 1 for phong, 2 for gouraud
uniform Material uMat;
uniform vec3 dirLight;
uniform vec3 viewPos;

varying vec3 vColor;
varying vec4 vNormal, vPos;
varying vec2 vTexCoord;

void main() {
   vec4 normal, light, view, reflection;
   vec3 D, S, A;
   float nDotL, vDotR;

   normal = normalize(vec4(vNormal.x, vNormal.y, vNormal.z, 0));
   view = normalize(uViewMatrix * uModelMatrix * vec4(viewPos.x, viewPos.y, viewPos.z, 1) - vPos);

   if (uLightMethod == 1) {
      light = normalize(uViewMatrix * vec4(dirLight, 0));
      nDotL = clamp(dot(light, normal), 0.0, 1.0);
      reflection = reflect(-light, normal);
      vDotR = clamp(dot(view, reflection), 0.0, 1.0);

      D = nDotL * uMat.dColor;
      S = pow(vDotR, uMat.shine) * uMat.sColor;
      A = uMat.aColor;

      gl_FragColor = vec4(D + S + A, 1);
   } else if (uLightMethod == 2) {
      vec4 texColor0 = vec4(vColor.x, vColor.y, vColor.z, 1);
      vec4 texColor1 = texture2D(uTexUnit, vTexCoord);
      gl_FragColor = vec4(texColor1[0], texColor1[1], texColor1[2], 1);
   } else {
      gl_FragColor = vec4(vColor.r, vColor.g, vColor.b, 1);
   }
}
