// Location data per vertex
attribute vec3 aPosition;
attribute vec3 aNormal;
attribute vec2 aTexCoord;

// Tranform matrices
uniform mat4 uProjMatrix;
uniform mat4 uViewMatrix;
uniform mat4 uModelMatrix;

// Lighting flag 0 for normals, 1 for phong, 2 for gouraud
uniform int uLightMethod;

// Data to pass to fragment shader
varying vec3 vColor;
varying vec4 vNormal, vPos;
varying vec2 vTexCoord;

void main() {
   vec4 pos, normal;
   
   pos = uViewMatrix * uModelMatrix * vec4(aPosition.x, aPosition.y, aPosition.z, 1);
   gl_Position = uProjMatrix * pos;
   vPos = pos;

   normal = normalize(uViewMatrix * uModelMatrix * vec4(aNormal.x, aNormal.y, aNormal.z, 0));
   vNormal = normal;
   
   if (uLightMethod == 0) {
      // DRAW NORMALS AS COLORS
      vColor = vec3(vNormal.r, vNormal.g, vNormal.b);
   } else {
      // PHONG LIGHTING
      // in the fragment shader
   }
   vTexCoord = aTexCoord;
}
